@@ -43,16 +43,18 @@
 #include "world.h"
 #include "render.h"
 
-#define TEST_ASM 1    /* Test inline assembly language. */
+
 #define VIEW_WIDTH  (fb->fb_width)
 #define VIEW_HEIGHT (fb->fb_height)
 
-/* These macros convert 16.16 fixed point numbers to and from a 4.28 format.
+/* These macros convert 16.16 fixed point numbers to and from a 2.30 format.
 **   The extra fractional precision is needed when doing doing 1 / distance
 **   calculations for perspective.
 */
-#define TO_FIX_4_28(f)   ((f) << 12)
-#define FROM_FIX_4_28(f) ((f) >> 12) 
+#define TO_FIX_2_30(f)   ((f) << 14)
+#define FROM_FIX_2_30(f) ((f) >> 14) 
+#define TO_FIX_8_24(f)   ((f) << 8)
+#define FROM_FIX_8_24(f) ((f) >> 8) 
 #define MAX_WALL_EVENTS 30
 
 #define MAX_ERROR  SCREEN_WIDTH
@@ -79,6 +81,9 @@ typedef struct {
 typedef struct {
      Boolean is_back_view;
      Wall *wall;
+     Boolean visible;
+     fixed pstart1, pend1, pstart2, pend2;
+     fixed dpstart1, dpend1, dpstart2, dpend2;
      fixed z;
      fixed dz;
 } Active_wall;
@@ -110,7 +115,6 @@ static Boolean wall_obscured(Vertex *common, Vertex *v1, Vertex *v2);
 static int remove_events(Active_wall *active, int n_active, int column);
 static Wall_intersection *calc_wall_heights(Active_wall *active, int n_active,
 					    int column, fixed height,
-					    fixed Vx, fixed Vy,
 					    Wall_intersection *cur_slice);
 static void draw_walls(Wall_intersection *int_list, View *v);
 static void draw_floors(Wall_intersection *int_list, View *v);
@@ -131,17 +135,18 @@ static int *fb_rows = NULL;
 static View_constants view_constants;
 
 
+#if defined(__GNUC__) && defined(ARCH_i86)
+#include "slice-gas86.c"
+#else
 #include "slice.c"
+#endif
 
 
 void init_renderer(int fb_width, int fb_height)
 {
      int i;
 
-#if TEST_ASM
-     if (fb_width != SCREEN_WIDTH)
-	  fatal_error("Frame buffer width must be 320 pixels for testing.");
-#endif
+
      if (fb != NULL)
 	  wtfree(fb);
      if (start_events != NULL)
@@ -161,7 +166,7 @@ void init_renderer(int fb_width, int fb_height)
 
 
 #if PROFILE
-double current_time(void)
+static double current_time(void)
 {
      struct timeval tv;
      struct timezone tz;
@@ -174,15 +179,40 @@ double current_time(void)
 }
 #endif
 
+#ifdef ARCH_SUN
+static void *memmove(void *dest, void *src, size_t size)
+{
+     int *lsrc = (int *) src;
+     int *ldest = (int *) dest;
 
-extern void render(World *w, View *v)
+     size /= 4;
+     /* Here, we assume that size is a multiple of four.  That's the only
+     **   way in which memmove will be used.  This assumption is nasty,
+     **   but life is rough . . . I'm bitter about having to implement
+     **   memmove in the first place.
+     */
+     if (lsrc < ldest) {
+	  ldest += size - 1;
+	  lsrc += size - 1;
+	  while (size-- > 0)
+	       *ldest-- = *lsrc--;
+     } else {
+	  while (size-- > 0)
+	       *ldest++ = *lsrc++;
+     }
+
+     return dest;
+}
+#endif
+
+
+void render(World *w, View *v)
 {
 #if PROFILE
      static int fps_count = 0;
      static double total_time = 0.0;
      double start_time = current_time();
 #endif
-
      init_buffers();
      calc_view_constants(v, VIEW_WIDTH, VIEW_HEIGHT);
      transform_vertices(w, v);
@@ -203,7 +233,7 @@ extern void render(World *w, View *v)
 }
 
 
-void transform_vertices(World *w, View *view)
+static void transform_vertices(World *w, View *view)
 {
      Vertex *vertex;
      fixed view_sin, view_cos;
@@ -227,7 +257,7 @@ void transform_vertices(World *w, View *view)
 }
 
 
-void clip_walls(World *w, View *v)
+static void clip_walls(World *w, View *v)
 {
      Wall *wall = (Wall *) w->walls->table;
      int i;
@@ -338,8 +368,8 @@ void clip_walls(World *w, View *v)
 /* Add a wall to the event list--one event is added for the start of the
 **   wall, and another is added to mark the end of the wall.
 */
-void add_wall_events(View *v, Wall *wall,
-	      fixed x1, fixed px1, fixed x2, fixed px2)
+static void add_wall_events(View *v, Wall *wall,
+			    fixed x1, fixed px1, fixed x2, fixed px2)
 {
      int fb1, fb2;
      fixed z1, z2;
@@ -347,10 +377,10 @@ void add_wall_events(View *v, Wall *wall,
 
 
      /* convert to frame buffer coordinates */
-     fb1 = (FIXED_TO_INT(fixmul(fixdiv(px1, v->view_plane_size),
-				INT_TO_FIXED(VIEW_WIDTH))) + VIEW_WIDTH) >> 1;
-     fb2 = (FIXED_TO_INT(fixmul(fixdiv(px2, v->view_plane_size),
-				INT_TO_FIXED(VIEW_WIDTH))) + VIEW_WIDTH) >> 1;
+     px1 = fixdiv(px1, view_constants.screen_dx + 1);
+     px2 = fixdiv(px2, view_constants.screen_dx + 1);
+     fb1 = FIXED_TO_INT(px1) + (VIEW_WIDTH >> 1);
+     fb2 = FIXED_TO_INT(px2) + (VIEW_WIDTH >> 1);
 
      /* There's no need to deal with walls that start and end in the same
      **   screen column.  In a properly contructed world, we're guaranteed
@@ -359,14 +389,14 @@ void add_wall_events(View *v, Wall *wall,
      if (fb1 == fb2)
 	  return;
 
-     /* Here we use a 4.28 fixed point format.  The result of this calculation
+     /* Here we use a 2.30 fixed point format.  The result of this calculation
      **   is always between 1 and zero, as the distance can never be less
      **   than the view plane distance.  The extra fractional bits are
-     **   critical for the inverses.  Note that using 4.28 restricts the
-     **   size of the view plane to something less than 8.
+     **   critical for the inverses.  Note that using 2.30 restricts the
+     **   size of the view plane to something less than 2.
      */
-     z1 = fixdiv(TO_FIX_4_28(v->eye_distance), x1);
-     z2 = fixdiv(TO_FIX_4_28(v->eye_distance), x2);
+     z1 = fixdiv(TO_FIX_2_30(v->eye_distance), x1);
+     z2 = fixdiv(TO_FIX_2_30(v->eye_distance), x2);
 
      if (fb1 < fb2) {
 
@@ -398,12 +428,11 @@ void add_wall_events(View *v, Wall *wall,
 }
 
 
-void render_walls(World *w, View *v)
+static void render_walls(World *w, View *v)
 {
      static int last_wall_count = 0;
      static Active_wall *active;
      int column;
-     int i;
      int active_count = 0;
      fixed Vx, Vy, dVy;
      Wall_intersection *cur_slice;
@@ -437,7 +466,7 @@ void render_walls(World *w, View *v)
      cur_slice = intersections;
 
      for (column = 0; column < VIEW_WIDTH; column++) {
-
+	  Active_wall *current, *last;
 	  /* Keep track of distances of walls in the active list.  Notice that
 	  **   we're not actually tracking the distances of walls, but 
 	  **   1 / distance instead.  That's because we can linearly
@@ -445,14 +474,23 @@ void render_walls(World *w, View *v)
 	  **   use distance are really using 1 / distance (i.e. distance
 	  **   appears in the denominator.
 	  */
-	  for (i = 0; i < active_count; i++)
-	       active[i].z += active[i].dz;
-
+		    
 	  active_count = add_events(active, active_count, column);
 	  cur_slice = calc_wall_heights(active, active_count, column,
-					v->height, Vx, Vy, cur_slice);
+					v->height, cur_slice);
 	  active_count = remove_events(active, active_count, column);
 
+	  last = active + active_count - 1;
+	  for (current = active; current <= last; current++) {
+	       current->z += current->dz;
+	       if (current->visible) {
+		    current->pstart1 += current->dpstart1;
+		    current->pend1 += current->dpend1;
+		    current->pstart2 += current->dpstart2;
+		    current->pend2 += current->dpend2;
+	       }
+	  }
+
 	  Vy += dVy;
      }
 
@@ -473,7 +511,7 @@ void render_walls(World *w, View *v)
 **   wall is obscured by another, the visible wall must be placed
 **   in front in the list.
 */
-int add_events(Active_wall *active, int n_active, int column)
+static int add_events(Active_wall *active, int n_active, int column)
 {
      int i, j;
      Wall_start *event;
@@ -538,6 +576,7 @@ int add_events(Active_wall *active, int n_active, int column)
 	  active[j].wall = wall;
 	  active[j].z = z;
 	  active[j].dz = event->dz;
+	  active[j].visible = False;
 	  active[j].is_back_view = event->is_back_view;
 	  n_active++;
      }
@@ -553,7 +592,7 @@ int add_events(Active_wall *active, int n_active, int column)
 **   transformed coordinates of the vertices, so the view point and view
 **   direction need not be passed explicitly.
 */
-Boolean wall_obscured(Vertex *common, Vertex *v1, Vertex *v2)
+static Boolean wall_obscured(Vertex *common, Vertex *v1, Vertex *v2)
 {
      fixed x1, y1, x2, y2;
      unsigned int sign1, sign2;
@@ -590,7 +629,7 @@ Boolean wall_obscured(Vertex *common, Vertex *v1, Vertex *v2)
 /* Remove walls from the active list.  Return the number of remaining
 ** active walls.
 */
-int remove_events(Active_wall *active, int n_active, int column)
+static int remove_events(Active_wall *active, int n_active, int column)
 {
      int i, j;
      Wall *wall;
@@ -610,22 +649,41 @@ int remove_events(Active_wall *active, int n_active, int column)
 }
 
 
-Wall_intersection *calc_wall_heights(Active_wall *active, int n_active,
-				     int column, fixed height,
-				     fixed Vx, fixed Vy,
-				     Wall_intersection *cur_slice)
+static Wall_intersection *calc_wall_heights(Active_wall *active, int n_active,
+					    int column, fixed height,
+					    Wall_intersection *cur_slice)
 {
      fixed top = FIXED_ONE_HALF, bottom = -FIXED_ONE_HALF;
 
 
      while (n_active-- > 0 && bottom < top) {
 	  Wall *wall = active->wall;
-	  fixed z = FROM_FIX_4_28(active->z);
+	  fixed z = active->z;
+	  fixed dz = active->dz;
 
-	  cur_slice->pstart1 = fixmul(wall->front->floor - height, z);
-	  cur_slice->pend1 = fixmul(wall->back->floor - height, z);
-	  cur_slice->pstart2 = fixmul(wall->back->ceiling - height, z);
-	  cur_slice->pend2 = fixmul(wall->front->ceiling - height, z);
+	  if (!active->visible) {
+	       active->pstart1 =
+		    fixmul2_30(TO_FIX_8_24(wall->front->floor - height), z);
+	       active->pend1 =
+		    fixmul2_30(TO_FIX_8_24(wall->back->floor - height), z);
+	       active->pstart2 =
+		    fixmul2_30(TO_FIX_8_24(wall->back->ceiling - height), z);
+	       active->pend2 =
+		    fixmul2_30(TO_FIX_8_24(wall->front->ceiling - height), z);
+	       active->dpstart1 =
+		    fixmul2_30(TO_FIX_8_24(wall->front->floor - height), dz);
+	       active->dpend1 =
+		    fixmul2_30(TO_FIX_8_24(wall->back->floor - height), dz);
+	       active->dpstart2 =
+		    fixmul2_30(TO_FIX_8_24(wall->back->ceiling - height), dz);
+	       active->dpend2 =
+		    fixmul2_30(TO_FIX_8_24(wall->front->ceiling - height), dz);
+	       active->visible = True;
+	  }
+	  cur_slice->pstart1 = FROM_FIX_8_24(active->pstart1);
+	  cur_slice->pend1 = FROM_FIX_8_24(active->pend1);
+	  cur_slice->pstart2 = FROM_FIX_8_24(active->pstart2);
+	  cur_slice->pend2 = FROM_FIX_8_24(active->pend2);
 
 	  if (active->is_back_view) {
 	       cur_slice->front = wall->back;
@@ -640,7 +698,7 @@ Wall_intersection *calc_wall_heights(Active_wall *active, int n_active,
 	  cur_slice->top = top;
 	  cur_slice->bottom = bottom;
 	  cur_slice->wall = wall;
-	  cur_slice->z = z;
+	  cur_slice->z = FROM_FIX_2_30(z);
 	  cur_slice->is_back_view = active->is_back_view;
 
 	  if (cur_slice->is_back_view) {
@@ -670,7 +728,7 @@ Wall_intersection *calc_wall_heights(Active_wall *active, int n_active,
 }
 
 
-void draw_walls(Wall_intersection *int_list, View *v)
+static void draw_walls(Wall_intersection *int_list, View *v)
 {
      int i = 0;
      int fb_column = VIEW_WIDTH - 1;
@@ -740,24 +798,46 @@ void draw_walls(Wall_intersection *int_list, View *v)
 	       continue;
 	  }
 
-	  t = wall_ray_intersection(Vx, Vy, wall);
-	  /* From t, calculate the integer coordinates in the texture
-	  **   bitmap.  For efficiency, we assume that the width of the
-	  **   texture is a power of two.
-	  */
-	  tex_column = FIXED_TO_INT(wall->xphase + fixmul(t, wall->xscale)) &
-	       (texture->width - 1);
+	  if (wall->sky) {
+	       fixed angle;
+
+	       /* Compute the angle of this column.  We'll use the angle
+	       **   exclusively to determine which texture column to
+	       **   display for this slice of sky--that's the way sky
+	       **   works.
+	       */
+	       angle = v->angle +
+		    fixdiv(FIXED_SCALE(v->arc, i - (VIEW_WIDTH >> 1)),
+			   INT_TO_FIXED(VIEW_WIDTH));
+	       angle -= FIXED_SCALE(FIXED_2PI, 
+				    FIXED_TO_INT(fixdiv(angle, FIXED_2PI)));
+	       angle = fixdiv(angle, FIXED_2PI);
+	       if (angle < FIXED_ZERO)
+		    angle = FIXED_ONE - angle;
+	       tex_column = FIXED_TO_INT(fixmul(angle, wall->xscale)) &
+		    (texture->width - 1);
+	       tex_dy = fixdiv(wall->yscale, INT_TO_FIXED(VIEW_HEIGHT));
+	  } else {
+	       t = wall_ray_intersection(Vx, Vy, wall);
+	       /* From t, calculate the integer coordinates in the texture
+		**   bitmap.  For efficiency, we assume that the width of the
+		**   texture is a power of two.
+		*/
+	       tex_column =
+		    FIXED_TO_INT(wall->xphase + fixmul(t, wall->xscale)) &
+			 (texture->width - 1);
+	       /* Test to avoid overflow here . . .  if the wall is so far
+		**   away that z (which is 1 / distance) is less than
+		**   FIXED_EPSILON, then it will be so small when rendered
+		**   that we can use a bogus value for tex_dy.
+		*/
+	       if (z < FIXED_EPSILON)
+		    tex_dy = 0;
+	       else
+		    tex_dy = fixdiv(wall->yscale, FIXED_SCALE(z, VIEW_HEIGHT));
+	  }
 	  tex_mask = texture->height - 1;
 	  tex_base = texture->texels + (tex_column << texture->log2width);
-	  /* Test to avoid overflow here . . .  if the wall is so far
-	  **   away that z (which is 1 / distance) is less than
-	  **   FIXED_EPSILON, then it will be so small when rendered
-	  **   that we can use a bogus value for tex_dy.
-	  */
-	  if (z < FIXED_EPSILON)
-	       tex_dy = 0;
-	  else
-	       tex_dy = fixdiv(wall->yscale, FIXED_SCALE(z, VIEW_HEIGHT));
 
 	  /* Now we can actually draw the walls, starting with
 	  **   the floor segment.
@@ -782,7 +862,11 @@ void draw_walls(Wall_intersection *int_list, View *v)
 	       fb_end = (VIEW_HEIGHT >> 1) - fb_end;
 	       fb_byte = fb->pixels + fb_column + fb_rows[fb_start];
 	       last_byte = fb->pixels + fb_column + fb_rows[fb_end];
-	       tex_y = fixmul(start1, wall->yscale) + wall->yphase;
+	       if (wall->sky)
+		    tex_y = fixmul(view_constants.row_view[fb_start],
+				   wall->yscale);
+	       else
+		    tex_y = fixmul(start1, wall->yscale) + wall->yphase;
 
 	       draw_wall_slice(fb_byte, last_byte, tex_base,
 			       tex_y, tex_dy, VIEW_WIDTH, texture->height,
@@ -810,7 +894,10 @@ void draw_walls(Wall_intersection *int_list, View *v)
 	       fb_end = (VIEW_HEIGHT >> 1) - fb_end;
 	       fb_byte = fb->pixels + fb_column + fb_rows[fb_start];
 	       last_byte = fb->pixels + fb_column + fb_rows[fb_end];
-	       tex_y = fixmul(start2, wall->yscale) + wall->yphase;
+	       if (wall->sky)
+		    tex_y = view_constants.row_view[fb_start];
+	       else
+		    tex_y = fixmul(start2, wall->yscale) + wall->yphase;
 
 	       draw_wall_slice(fb_byte, last_byte, tex_base,
 			       tex_y, tex_dy, VIEW_WIDTH, texture->height,
@@ -822,18 +909,19 @@ void draw_walls(Wall_intersection *int_list, View *v)
 }
 
 
-void draw_floors(Wall_intersection *int_list, View *v)
+static void draw_floors(Wall_intersection *int_list, View *v)
 {
      int i;
      Wall_intersection *last, *current;
      Wall_intersection *save_current, *save_last;
      fixed start, end;
+     fixed max_floor, min_ceiling;
+     fixed last_max_floor = FIXED_ZERO, last_min_ceiling = FIXED_ZERO;
 
 
      last = current = int_list;
 
      for (i = 1; i < VIEW_WIDTH; i++) {
-	  fixed max_floor, min_ceiling;
 
 	  for (; current->wall != NULL; current++);
 	  save_current = ++current;
@@ -893,15 +981,18 @@ void draw_floors(Wall_intersection *int_list, View *v)
 		    current++;
 	  }
 
+	  /* draw ceiling slices generated by floor walls */
 	  while (last->wall != NULL) {
-	       start = last->pend2;
+	       start = MAX(last_max_floor, last->pend2);
 	       end = MIN(max_floor, last->top);
 	       if (start < end)
 		    draw_ceiling_slices(last->front, start, end, v, i - 1);
 	       last++;
 	  }
-
 	  last = save_current;
+
+	  last_max_floor = max_floor;
+	  last_min_ceiling = min_ceiling;
      }
 
      save_last = last;
@@ -934,7 +1025,8 @@ void draw_floors(Wall_intersection *int_list, View *v)
 }
 
 
-void draw_floor_slices(Region *r, fixed start, fixed end, View *v, int column)
+static void draw_floor_slices(Region *r, fixed start, fixed end,
+			      View *v, int column)
 {
      int screen_column = VIEW_WIDTH - column - 1;
      int start_row, end_row;
@@ -984,8 +1076,8 @@ void draw_floor_slices(Region *r, fixed start, fixed end, View *v, int column)
 }
 
 
-void draw_ceiling_slices(Region *r, fixed start, fixed end,
-			 View *v, int column)
+static void draw_ceiling_slices(Region *r, fixed start, fixed end,
+				View *v, int column)
 {
      int screen_column = VIEW_WIDTH - column - 1;
      int start_row, end_row;
@@ -1037,9 +1129,9 @@ void draw_ceiling_slices(Region *r, fixed start, fixed end,
 
 /* Calculate the value of the parameter t at the intersection
 **   of the view ray and the wall.  t is 0 at the origin of
-**   the wall, and 1 at he other endpoint.
+**   the wall, and 1 at the other endpoint.
 */
-fixed wall_ray_intersection(fixed Vx, fixed Vy, Wall *wall)
+static fixed wall_ray_intersection(fixed Vx, fixed Vy, Wall *wall)
 {
      fixed denominator, Nx, Ny, Wx, Wy;
      
@@ -1062,7 +1154,7 @@ fixed wall_ray_intersection(fixed Vx, fixed Vy, Wall *wall)
 }
 
 
-void init_buffers(void)
+static void init_buffers(void)
 {
      int i;
 
@@ -1076,7 +1168,7 @@ void init_buffers(void)
 /* Calculate values that are dependent only on the screen dimensions and
 **   the view.
 */
-void calc_view_constants(View *v, int screen_width, int screen_height)
+static void calc_view_constants(View *v, int screen_width, int screen_height)
 {
      static int last_height = 0, last_width = 0;
      int i;
