@@ -20,27 +20,29 @@
 **  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
 
+/* GIF87a code from Thomas Hamren (d0hampe@dtek.chalmers.se) */
+
 #ifndef TEXTURE_H_
 #define TEXTURE_H_
 
+#define TEXTURE_MAGIC "wttex"
+#define TEXTURE_MAGIC_LENGTH 5
 
-#define TEXTURE_MAGIC 0x8866ed04
-
-typedef enum {
-     TEX_SOLID,
-     TEX_TRANSPARENT
-} Texture_type;
+#define GIF_MAGIC "GIF87"
 
 
 typedef struct {
-     Texture_type type;
-     unsigned char base_color;
      int width, height;
-     int log2width;     /* for efficiency, when width is a power of 2 */
+     int log2width;
      unsigned char *texels;
 } Texture;
 
-extern Texture *new_texture(Texture_type, int, int);
-extern Texture *read_texture_file(char *);
+typedef enum { WTTEX, GIF87a, Unknown } TexFormat;
+
+extern Texture *new_texture(int width, int height);
+extern Texture *read_texture_file(char *filename);
+
+/* The following function is located in gifload.c, not texture.c. */
+extern Texture *LoadGIF(FILE *fp, char *fname);
 
 #endif
