@@ -43,3 +43,25 @@ typedef enum { False, True } Boolean;
 #ifndef __GNUC__
 #define inline
 #endif
+
+/* For compilers without NULL defined in stdlib.h */
+#ifndef NULL
+#include <stdio.h>
+#endif
+
+/* More compatibility stuff . . . */
+#ifndef PATH_MAX
+#ifdef _POSIX_PATH_MAX
+#define PATH_MAX _POSIX_PATH_MAX
+#else
+#define PATH_MAX 1024
+#endif
+#endif
+
+#ifndef EXIT_FAILURE
+#define EXIT_FAILURE 1
+#endif
+#ifndef EXIT_SUCCESS
+#define EXIT_SUCCES 0
+#endif
+
